﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;

namespace JpegGps
{
    [Serializable]
    public class JpegGpsException : ApplicationException
    {
        public JpegGpsException() { }
        public JpegGpsException(string message) : base(message) { }
    }

    public class JpegGpsValue
    {
        public char Prefix { get; set; }
        public double Dd { get; set; }
        public double Mm { get; set; }
        public double Ss { get; set; }
        public override string ToString()
        {
            return string.Format("{0}{1}時{2}分{3}秒", Prefix, Dd, Mm, Ss);
        }
    }

    public class JpegGpsLocation
    {
        public JpegGpsValue Latitude = new JpegGpsValue();
        public JpegGpsValue Longtitude = new JpegGpsValue();
    };

    internal class JpegGpsWorker
    {
        private Dictionary<formatIds, format> formats;
        private endianGetter getter;
        private byte[] sec;

        private enum markers
        {
            SOI = 0xD8, SOS = 0xDA, EXIF = 0xE1,
        }
        private enum tags
        {
            GPS_LAT_REF = 1, GPS_LAT = 2, GPS_LONG_REF = 3,
            GPS_LONG = 4, GPS_ALT_REF = 5, GPS_ALT = 6,
            MAKER_NOTE = 0x927C, GPSINFO = 0x8825,
        }
        private abstract class endianGetter
        {
            protected byte[] sec;
            internal abstract int get16u(int idx);
            internal int get16s(int idx)
            {
                var val = get16u(idx);
                if (val <= short.MaxValue) return val;
                else return val - 65536;
            }
            internal tags get16tags(int idx)
            {
                return (tags)get16u(idx);
            }
            internal formatIds get16formats(int idx)
            {
                return (formatIds)get16u(idx);
            }
            internal abstract int get32s(int idx);
            internal uint get32u(int idx)
            {
                return (uint)get32s(idx);
            }
            internal double calcDen(int idx)
            {
                double den = get32s(4 + idx);
                return den == 0.0 ? 0.0 : get32s(idx) / den;
            }
            public endianGetter(byte[] exifSection)
            {
                this.sec = exifSection;
            }
        }
        private class BigEndianGetter : endianGetter
        {
            internal override int get16u(int idx)
            {
                return (sec[idx + 0] << 8) | sec[idx + 1];
            }
            internal override int get32s(int idx)
            {
                return (sec[idx + 0] << 24) | (sec[idx + 1] << 16)
                    | (sec[idx + 2] << 8) | (sec[idx + 3] << 0);
            }
            public BigEndianGetter(byte[] exifSection) : base(exifSection) { }
        }
        private class LittleEndianGetter : endianGetter
        {
            internal override int get16u(int idx)
            {
                return (sec[idx + 1] << 8) | sec[idx + 0];
            }
            internal override int get32s(int idx)
            {
                return (sec[idx + 3] << 24) | (sec[idx + 2] << 16)
                    | (sec[idx + 1] << 8) | (sec[idx + 0] << 0);
            }
            public LittleEndianGetter(byte[] exifSection) : base(exifSection) { }
        }

        private int calcValueOffset(int offsetBase, uint exifLength, int dirEntryOffset, tags tag, int byteCount)
        {
            uint OffsetVal = getter.get32u(dirEntryOffset + 8);
            if (OffsetVal + byteCount > exifLength) throw new JpegGpsException(string.Format("値のポインターが不正です。tag: {0}", tag));
            return offsetBase + (int)OffsetVal;
        }

        private enum formatIds
        {
            BYTE = 1, STRING = 2, USHORT = 3, ULONG = 4,
            URATIONAL = 5, SBYTE = 6, UNDEFINED = 7,
            SSHORT = 8, SLONG = 9, SRATIONAL = 10,
            SINGLE = 11, DOUBLE = 12,
        }
        private class format
        {
            internal int bytes;
            internal Func<byte[], int, double> getter;
            public format(int bytes, Func<byte[], int, double> getter)
            {
                this.bytes = bytes;
                this.getter = getter;
            }
        }

        private format getFormat(formatIds fmt, string supply)
        {
            format r;
            if (formats.TryGetValue(fmt, out r)) return r;
            throw new JpegGpsException(fmt + "は処理できないフォーマットです。" + supply);
        }

        private void get3Items(formatIds format, int componentSize, int valuePtr, JpegGpsValue val)
        {
            var fmt = getFormat(format, "Exifヘッダー");
            val.Dd = fmt.getter(sec, valuePtr);
            val.Mm = fmt.getter(sec, valuePtr + componentSize);
            val.Ss = fmt.getter(sec, valuePtr + componentSize * 2);
        }
        private void getLetter(int valuePtr, JpegGpsValue val)
        {
            val.Prefix = (char)sec[valuePtr];
        }
        private JpegGpsLocation processGpsInfo(int offsetDirStart, int offsetBase, uint ExifLength)
        {
            JpegGpsLocation Location = new JpegGpsLocation();
            foreach (var dirEntryOffset in Enumerable.Range(0, getter.get16u(offsetDirStart))
                .Select(c => offsetDirStart + 2 + c * 12))
            {
                if (dirEntryOffset + 12 > offsetBase + ExifLength) throw new JpegGpsException("GPS情報のディレクトリはExifの終わりを超えています");

                tags tag = getter.get16tags(dirEntryOffset);
                formatIds format = getter.get16formats(dirEntryOffset + 2);
                int componentSize = getFormat(format, "Exif GPS tag: " + tag).bytes;
                int byteCount = getter.get32s(dirEntryOffset + 4) * componentSize;
                int valuePtr = byteCount > 4 ? calcValueOffset(offsetBase, ExifLength, dirEntryOffset, tag, byteCount) : dirEntryOffset + 8;
                switch (tag)
                {
                    case tags.GPS_LAT_REF:
                        getLetter(valuePtr, Location.Latitude);
                        break;
                    case tags.GPS_LONG_REF:
                        getLetter(valuePtr, Location.Longtitude);
                        break;
                    case tags.GPS_LAT:
                        get3Items(format, componentSize, valuePtr, Location.Latitude);
                        break;
                    case tags.GPS_LONG:
                        get3Items(format, componentSize, valuePtr, Location.Longtitude);
                        break;
                }
            }
            return Location;
        }

        private JpegGpsLocation processTagGpsInfo(int offsetBase, uint exifLength, int byteCount, int valuePtr)
        {
            int SubdirStartOffset = offsetBase + getter.get32s(valuePtr);
            if (SubdirStartOffset < offsetBase || SubdirStartOffset > offsetBase + exifLength)
                throw new JpegGpsException("不正なGPSディレクトリリンクです");
            else return processGpsInfo(SubdirStartOffset, offsetBase, exifLength);
        }

        private JpegGpsLocation processExifDir(int offsetDirStart, uint exifLength)
        {
            foreach (var dirEntryOffset in Enumerable.Range(0, getter.get16u(offsetDirStart))
                .Select(c => offsetDirStart + 2 + c * 12))
            {
                tags tag = getter.get16tags(dirEntryOffset);
                if (tag == tags.MAKER_NOTE)
                    throw new JpegGpsException("Maker noteを検出しました");

                uint numberOfComponents = getter.get32u(dirEntryOffset + 4);

                if (numberOfComponents > 0x10000)
                    throw new JpegGpsException(string.Format("コンポーネントが多すぎます。数:{0} tag: {1}", numberOfComponents, tag));

                int byteCount = (int)numberOfComponents
                    * getFormat(getter.get16formats(dirEntryOffset + 2), "tag " + tag + " in Exif").bytes;

                int valuePtr = byteCount > 4
                    ? calcValueOffset(8, exifLength, dirEntryOffset, tag, byteCount)
                    : dirEntryOffset + 8;
                if (tag == tags.GPSINFO) return processTagGpsInfo(8, exifLength, byteCount, valuePtr);
            }
            return null;
        }

        private bool isAlignmentMarker(char c)
        {
            return sec[8] == c && sec[9] == c;
        }

        private JpegGpsLocation processExif()
        {
            bool isBigEndian = isAlignmentMarker('M');
            getter = isBigEndian ? (endianGetter)new BigEndianGetter(sec) : (endianGetter)new LittleEndianGetter(sec);

            if (isBigEndian == false && !isAlignmentMarker('I')) throw new JpegGpsException("不正なアラインメントマーカーです。");
            if (getter.get16u(10) != 0x2a) throw new JpegGpsException("不正なExifのスタートです。");
            return processExifDir(8 + getter.get32s(12), (uint)sec.Length - 8);
        }

        private static markers seekMarker(Stream infile)
        {
            int prev = 0, marker = 0;
            for (; ; )
            {
                marker = infile.ReadByte();
                if (marker != 0xff && prev == 0xff) return (markers)marker;
                prev = marker;
            }
        }

        internal JpegGpsLocation ReadJpegSections(Stream infile)
        {
            if (infile.ReadByte() != 0xff
                || infile.ReadByte() != (int)markers.SOI) return null;
            for (; ; )
            {
                markers marker = seekMarker(infile);
                var h = infile.ReadByte();
                var l = infile.ReadByte();
                var length = (h << 8) | l;
                if (length < 2) throw new JpegGpsException("不正なマーカーです。");

                sec = new byte[length];

                if (infile.Read(sec, 2, length - 2) != length - 2) throw new JpegGpsException("ファイルがもうありません");

                if (marker == markers.SOS) return null;
                else if (marker == markers.EXIF &&
                    sec.Skip(2).Take(6).Select(c => (char)c).SequenceEqual("Exif\0\0")) return processExif();
            }
        }

        public JpegGpsWorker()
        {
            formats = new Dictionary<formatIds, format>()
            {
                {formatIds.BYTE, new format(1,(sec,idx)=>(double)sec[idx])},
                {formatIds.STRING, new format(1,null)},
                {formatIds.USHORT, new format(2,(sec,idx)=>getter.get16u( idx))},
                {formatIds.ULONG, new format(4,(sec,idx)=>(double)getter.get32u(idx))},
                {formatIds.URATIONAL, new format(8,(sec,idx)=>(double)getter.calcDen(idx))},
                {formatIds.SBYTE, new format(1,(sec,idx)=>(double)((sbyte)sec[idx]))},
                {formatIds.UNDEFINED, new format(1,null)},
                {formatIds.SSHORT , new format(2,(sec,idx)=>(double)getter.get16s(idx))},
                {formatIds.SLONG , new format(4,(sec,idx)=>(double)getter.get32s(idx))},
                {formatIds.SRATIONAL, new format(8,(sec,idx)=>(double)getter.calcDen(idx))},
                {formatIds.SINGLE, new format(4,null)},
                {formatIds.DOUBLE, new format(8,null)},
            };
        }
    }
    public static class JpegGps
    {
        public static JpegGpsLocation ProcessFile(Stream stream)
        {
            try
            {
                return new JpegGpsWorker().ReadJpegSections(stream);
            }
            finally
            {
                stream.Close();
            }
        }
    }
}
